// runnerline_print.cpp
// Revision 3-jan-2007

#include "runnerline_impl.h"

#include "using.h"
#include "sysvar.h"
#include "util.h"
using util::to_string;

#include "trace.h"


namespace sysvar= blassic::sysvar;
using namespace blassic::file;


namespace {


// Class to make transitory modifications in the status of a channel
// and restore it on exit.

class PrintChannel {
	BlFile & bf;
	bool inversesaved;
	bool inversevalue;
	bool inksaved;
	int inkvalue;
	bool papersaved;
	int papervalue;
	bool brightsaved;
	bool brightvalue;
public:
	PrintChannel (BlFile & bf);
	~PrintChannel ();

	BlFile & file ();
	void endline ();
	void tab ();
	void tab (size_t n);
	void putspaces (size_t n);
	void gotoxy (int x, int y);

	void inverse (bool inv);
	void ink (int color);
	void paper (int color);
	void bright (bool br);

	void operator << (const std::string & s);
	void operator << (char c);
	void operator << (BlNumber n);
	void operator << (BlInteger n);
};


PrintChannel::PrintChannel (BlFile & bf) :
	bf (bf),
	inversesaved (false),
	inversevalue (0),
	inksaved (false),
	inkvalue (0),
	papersaved (false),
	papervalue (0),
	brightsaved (false),
	brightvalue (0)
{
	TRF;
}

PrintChannel::~PrintChannel ()
{
	TRF;

	TRDEB ("Flushing channel");
	bf.flush ();

	if (inversesaved)
		bf.inverse (inversevalue);
	if (brightsaved)
		bf.bright (brightvalue);
	if (inksaved)
		bf.setcolor (inkvalue);
	if (papersaved)
		bf.setbackground (papervalue);
}

BlFile & PrintChannel::file ()
{
	return bf;
}

void PrintChannel::endline ()
{
	bf.endline ();
}

void PrintChannel::tab ()
{
	bf.tab ();
}

void PrintChannel::tab (size_t n)
{
	bf.tab (n);
}

void PrintChannel::putspaces (size_t n)
{
	bf.putspaces (n);
}

void PrintChannel::gotoxy (int x, int y)
{
	bf.flush ();
	bf.gotoxy (x, y);
}

void PrintChannel::inverse (bool inv)
{
	if (! inversesaved)
	{
		inversevalue= bf.getinverse ();
		inversesaved= true;
	}
	bf.flush ();
	bf.inverse (inv);
}

void PrintChannel::ink (int color)
{
	if (! inksaved)
	{
		inkvalue= bf.getcolor ();
		inksaved= true;
	}
	bf.flush ();
	bf.setcolor (color);
}

void PrintChannel::paper (int color)
{
	if (! papersaved)
	{
		papervalue= bf.getbackground ();
		papersaved= true;
	}
	bf.flush ();
	bf.setbackground (color);
}

void PrintChannel::bright (bool br)
{
	if (! brightsaved)
	{
		brightvalue= bf.getbright ();
		brightsaved= true;
	}
	bf.flush ();
	bf.bright (br);
}

void PrintChannel::operator << (const std::string & s)
{
	bf << s;
}

void PrintChannel::operator << (char c)
{
	bf << c;
}

void PrintChannel::operator << (BlNumber n)
{
	bf << n;
}

void PrintChannel::operator << (BlInteger n)
{
	bf << n;
}


} // namespace


#define requiretoken(c) if (token.code == c) ; else throw ErrSyntax

#define expecttoken(c) \
	do { \
		gettoken (); \
		if (token.code != c) throw ErrSyntax; \
	} while (0)


using namespace blassic::vartype;


void RunnerLineImpl::print_using (BlFile & out)
{
	TRF;

	using blassic::VectorUsing;
	using blassic::Using;

	std::string format= expectstring ();
	VectorUsing usingf;
	parseusing (format, usingf);
	if (token.code == keyComma || token.code == ';')
		gettoken ();
	const size_t l= usingf.size ();
	size_t ind= 0;
	Using * pf= usingf [ind];
	for (;;)
	{
		if (ind == 0 && pf->isliteral () )
		{
			pf->putliteral (out);
			ind= (ind + 1) % l;
			if (ind == 0)
				throw ErrFunctionCall;
			pf= usingf [ind];
		}
		BlResult result;
		eval (result);
		switch (result.type () )
		{
		case VarNumber:
			pf->putnumeric (out, result.number () );
			break;
		case VarInteger:
			pf->putnumeric (out, result.number () );
			break;
		case VarString:
			pf->putstring (out, result.str () );
			break;
		default:
			ASSERT (false);
			throw ErrBlassicInternal;
		}
		ind= (ind + 1) % l;
		pf= usingf [ind];
		if (ind != 0 && pf->isliteral () )
		{
			pf->putliteral (out);
			ind= (ind + 1) % l;
			// Seen unnecessary, and is erroneous.
			//if (ind == 0)
			//	throw ErrFunctionCall;
			pf= usingf [ind];
		}
		if (endsentence () )
		{
			out.endline ();
			break;
		}
		if (token.code == ';' || token.code == keyComma)
		{
			gettoken ();
			if (endsentence () )
				break;
		}
	}
}

bool RunnerLineImpl::do_PRINT ()
{
	TRF;

	// Same function used for PRINT and LPRINT, differing only
	// in the default channel.
	BlChannel channel= (token.code == keyLPRINT) ?
		PrinterChannel : DefaultChannel;
	gettoken ();
	if (token.code == '#')
	{
		channel= expectchannel ();
		// Allow ';' for Spectrum compatibility.
		if (token.code == keyComma || token.code == ';')
			gettoken ();
		else
			require_endsentence ();
	}

	TRDEBS ("Channel: " << channel);

	//BlFile & out= getfile (channel);
	//TRDEBS ("Text window: " << out.istextwindow () );
	//TRDEBS ("File: " << out.isfile () );
	//PrintChannel channelsave (out);
	PrintChannel out (getfile (channel) );

	if (token.code == '@')
	{
		BlResult result;
		expect (result);
		BlInteger pos= result.integer ();
		requiretoken (keyComma);
		gettoken ();
		#if 0
		BlInteger row= (pos / 32) + 1;
		BlInteger col= (pos % 32) + 1;
		if (graphics::ingraphicsmode () )
			graphics::locate (row, col);
		else
			locate (row, col);
		#else
		out.gotoxy (pos % 32, pos / 32);
		#endif
	}
	if (endsentence () )
	{
		//out << '\n';
		out.endline ();
		return false;
	}

	bool spacebeforenumber=
		sysvar::hasFlags1 (sysvar::SpaceBefore);
	BlResult result;
	size_t n;
	bool ended= false;
	do {
		switch (token.code) {
		case keyComma:
		case ';':
			// Separators can be before any statement,
			break;
		case keyUSING:
			print_using (out.file () );
			ended= true;
			break;
		case keyTAB:
			//getparenarg (result);
			// Not required to improve Sinclair ZX compatibility.
			expect (result);
			n= result.integer ();

			if (! sysvar::hasFlags1 (sysvar::TabStyle) )
			{
				if (n > 0)
					--n;
			}

			TRDEBS ("TAB " << n);
			out.tab (n);
			break;
		case keySPC:
			getparenarg (result);
			n= result.integer ();
			out.putspaces (n);
			break;
		case keyAT:
			{
				expect (result);
				BlInteger row= result.integer ();
				requiretoken (keyComma);
				expect (result);
				BlInteger col= result.integer ();
				out.gotoxy (col, row);
			}
			break;
		case keyINK:
			out.ink (expectinteger () );
			break;
		case keyPAPER:
			out.paper (expectinteger () );
			break;
		case keyINVERSE:
			out.inverse (expectinteger () );
			break;
		case keyBRIGHT:
			out.bright (expectinteger () );
			break;
		default:
			eval (result);
			switch (result.type () ) {
			case VarString:
				{
					TRDEB ("string");
					std::string txt= result.str ();
					TRDEB (txt);
					//out << result.str ();
					out << txt;
				}
				break;
			case VarNumber:
				{
					BlNumber n= result.number ();
					if (spacebeforenumber && n >= 0)
						out << ' ';
					out << n;
				}
				break;
			case VarInteger:
				{
					BlInteger n= result.integer ();
					if (spacebeforenumber && n >= 0)
						out << ' ';
					out << n;
				}
				break;
			default:
				ASSERT (false);
				throw ErrBlassicInternal;
			}
		}
		if (ended)
			break;
		if (endsentence () )
		{
			out.endline ();
			break;
		}
		#if 0
		if (token.code != ';' && token.code != keyComma)
			throw ErrSyntax;
		if (token.code == keyComma)
			out.tab ();
		gettoken ();
		#else
		// Now separator is not required
		switch (token.code)
		{
		case keyComma:
			out.tab ();
			gettoken ();
			break;
		case ';':
			gettoken ();
			break;
		default:
			; // Nothing
		}
		#endif
	} while (! endsentence () );
	return false;
}

// End of runnerline_print.cpp
